unit FrmAutomation;

{

   COPYRIGHT 2017 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    Oogstplein 6
    7545 HP Enschede
    the Netherlands
    http://www.iaf.nl/Users/BlueHell/
    http://bluehell.electro-music.com/
    j_dot.punter2@t2iaf_dot.nl

  All rights attributed to Blue Hell are owned by Jan Punter.
}

interface

uses

  Winapi.Windows, Winapi.Messages,

  System.SysUtils, System.Variants, System.Classes, System.Math, System.IniFiles, System.Types,
  System.Generics.Defaults, System.Generics.Collections,

  Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.ExtCtrls, Vcl.StdCtrls, Vcl.Buttons,

  Globals, knobs2013, KnobsUtils, Vcl.Samples.Spin, Vcl.ComCtrls, KnobsConversions, ComplexMath;


type

  TScrollBox = class( Vcl.Forms.TScrollBox)
  protected
    procedure   CreateParams( var Params: TCreateParams);                                                      override;
  end;


  TKnobsTimeIndexList = record
  private
    FTime     : TKnobsTimeStamp;
    FDuration : TKnobsTimeStamp;
    FItems    : TArray<Integer>;
  public
    procedure   Create( aTime: TKnobsTimeStamp);
    procedure   Clear;
    procedure   AddItem( anItem: Integer);
    function    Count : Integer;
    procedure   SetDuration( aValue: TKnobsTimeStamp);
    function    Time     : TKnobsTimeStamp;
    function    Duration : TKnobsTimeStamp;
    function    Item( anIndex: Integer): Integer;
  end;


  TKnobsTimeIndexLists = TArray<TKnobsTimeIndexList>;


  TKnobsTimeIndex = class
  private
    FData : TKnobsTimeIndexLists;
  private
    function    GetCount     : Integer;
    function    GetStartTime : TKnobsTimeStamp;
    function    GetTimeSpan  : TKnobsTimeStamp;
    function    GetTime      ( anIndex: Integer) : TKnobsTimeStamp;
    function    GetDuration  ( anIndex: Integer) : TKnobsTimeStamp;
    function    GetItemCount ( anIndex: Integer) : Integer;
    function    GetItem      ( i, j   : Integer) : Integer;
    procedure   Sort;
  public
    procedure   AddItem( const aTimeIndexList: TKnobsTimeIndexList);                                           overload;
    procedure   AddItem( aTime: TKnobsTimeStamp; anIndex: Integer);                                            overload;
    procedure   Clear;
    procedure   SetFromHistory( const aData: TKnobsEditHistory);
    function    FindTime( aTime: TKnobsTimeStamp): Integer;
  public
    property    Count                        : Integer            read GetCount;
    property    StartTime                    : TKnobsTimeStamp    read GetStartTime;
    property    TimeSpan                     : TKnobsTimeStamp    read GetTimeSpan;
    property    Time      [ anIndex: Integer]: TKnobsTimeStamp    read GetTime;
    property    Duration  [ anIndex: Integer]: TKnobsTimeStamp    read GetDuration;
    property    ItemCount [ anIndex: Integer]: Integer            read GetItemCount;
    property    Item      [ i, j   : Integer]: Integer            read GetItem;
  end;


  TKnobsEditDataTimeList = class
  private
    FItems : TArray<PKnobsEditData>;
  private
    function    GetCount: Integer;
    function    GetTime( anIndex: Integer): TKnobsTimeStamp;
    function    GetItem( anIndex: Integer): PKnobsEditData;
  public
    constructor Create;
    destructor  Destroy;                                                                                       override;
    procedure   AddItem( anItem: PKnobsEditData);
    function    FindTime( aTime: TKnobsTimeStamp): Integer;
    procedure   FillGrid( const aGrid: TKnobsDataMaker; aStartTime, aTimeSpan: TKnobsTimeStamp);
  public
    property    Count                   : Integer         read GetCount;
    property    Time[ anIndex: Integer] : TKnobsTimeStamp read GetTime;
    property    Item[ anIndex: Integer] : PKnobsEditData  read GetItem;
  end;


  TKnobsEditDataTimeIndex = class
  private
    FData      : TArray<TKnobsEditDataTimeList>;
    FPathIndex : TStringList;
  private
    function    GetCount                                       : Integer;
    function    GetData     ( anIndex: Integer)                : TKnobsEditDataTimeList;
    function    GetTimeCount( anIndex: Integer)                : Integer;
    function    GetTime     ( aDataIndex, aTimeIndex: Integer) : TKnobsTimeStamp;
  public
    constructor Create;
    destructor  Destroy;                                                                                       override;
    procedure   Clear;
    procedure   SetFromHistory( const aData: TKnobsEditHistory);
    function    FindPath( const aPath: string): Integer;
    procedure   FillGrid( anIndex: Integer; const aGrid: TKnobsDataMaker; aStartTime, aTimeSpan: TKnobsTimeStamp);
  public
    property    Count                                       : Integer                read GetCount;
    property    Data     [ anIndex: Integer               ] : TKnobsEditDataTimeList read GetData;
    property    TimeCount[ anIndex: Integer               ] : Integer                read GetTimeCount;
    property    Time     [ aDataIndex, aTimeIndex: Integer] : TKnobsTimeStamp        read GetTime;
  end;


  TFormAutomation = class(TForm)
    PanelTop: TPanel;
    PanelAutomationGrid: TPanel;
    PanelBottom: TPanel;
    ScrollBoxAutomationGrid: TScrollBox;
    SpeedButtonPlay: TKnobsTimerSpeedButton;
    SpeedButtonPause: TKnobsTimerSpeedButton;
    SpeedButtonLeftStop: TKnobsTimerSpeedButton;
    SpeedButtonRightStop: TKnobsTimerSpeedButton;
    SpeedButtonLeftFast: TKnobsTimerSpeedButton;
    SpeedButtonRightFast: TKnobsTimerSpeedButton;
    SpeedButtonRecord: TKnobsTimerSpeedButton;
    KnobsAutomationGrid: TKnobsAutomationGrid;
    SpeedButtonLoop: TKnobsTimerSpeedButton;
    KnobsSelectorSnap: TKnobsSelector;
    KnobsLedAutomationActive: TKnobsLed;
    TimerStatusUpdate: TTimer;
    KnobsSelectorClear: TKnobsSelector;
    LabelZoom: TLabel;
    SpinButtonZoom: TSpinButton;
    LabelAutomationInfo: TLabel;
    EditDummy: TEdit;
    PanelDetails: TPanel;
    SplitterDetails: TSplitter;
    PanelDetailsTop: TPanel;
    LabelTime: TLabel;
    ListView: TListView;
    PanelDetailsEditor: TPanel;
    SplitterBottom: TSplitter;
    BitBtnToggleNames: TBitBtn;
    Panel1: TPanel;
    Panel2: TPanel;
    KnobsDataMaker: TKnobsDataMaker;
    Label43: TLabel;
    EditAutomationSize: TEdit;
    LabelDebug: TLabel;
    datagraph_splines: TKnobsSelector;
    BitBtnSaveRecording: TBitBtn;
    BitBtnLoadRecording: TBitBtn;
    KnobsSelectorSend: TKnobsSelector;
    CheckBoxLockValue: TCheckBox;
    CheckBoxLockTime: TCheckBox;
    TimerBlink: TTimer;
    procedure SpeedButtonRecordClick(Sender: TObject);
    procedure SpeedButtonLoopClick(Sender: TObject);
    procedure SpeedButtonStopClick(Sender: TObject);
    procedure SpeedButtonPlayClick(Sender: TObject);
    procedure SpeedButtonPauseClick(Sender: TObject);
    procedure SpeedButtonLeftStopClick(Sender: TObject);
    procedure SpeedButtonRightStopClick(Sender: TObject);
    procedure SpeedButtonLeftFastClick(Sender: TObject);
    procedure SpeedButtonRightFastClick(Sender: TObject);
    procedure TimerStatusUpdateTimer(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure KnobsSelectorSnapValueChanged(const aSender: TObject; const aPath, aControlType: string;
      aValue: Double; IsFinal, IsAutomation: Boolean);
    procedure KnobsSelectorClearValueChanged(const aSender: TObject; const aPath, aControlType: string; aValue: Double;
      IsFinal, IsAutomation: Boolean);
    procedure SpinButtonZoomDownClick(Sender: TObject);
    procedure SpinButtonZoomUpClick(Sender: TObject);
    procedure FormMouseWheel(Sender: TObject; Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint;
      var Handled: Boolean);
    procedure KnobsLedAutomationActiveClick(Sender: TObject);
    procedure ControlUnfocus(const aSender: TObject);
    procedure FormResize(Sender: TObject);
    procedure EditDummyChange(Sender: TObject);
    procedure KnobsAutomationGridClick(const aSender: TObject; aButton: TMouseButton; aShift: TShiftState; anX,
      anY: Integer);
    procedure EditAutomationSizeChange(Sender: TObject);
    procedure ListViewSelectItem(Sender: TObject; Item: TListItem; Selected: Boolean);
    procedure KnobsKnobKnobPositionChanged(const aSender: TKnobsValuedControl; const aPath, aControlType: string;
      aPosition: Integer);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure BitBtnToggleNamesClick(Sender: TObject);
    procedure datagraph_splinesValueChanged(const aSender: TObject; const aPath, aControlType: string; aValue: Double;
      IsFinal, IsAutomation: Boolean);
    procedure BitBtnSaveRecordingClick(Sender: TObject);
    procedure BitBtnLoadRecordingClick(Sender: TObject);
    procedure KnobsSelectorSendValueChanged(const aSender: TObject; const aPath, aControlType: string; aValue: Double;
      IsFinal, IsAutomation: Boolean);
    procedure KnobsDataMakerPointChanged(const aSender: TObject; const aPath: string; anIndex: Integer; anX, anY: Double);
    procedure CheckBoxLockValueClick(Sender: TObject);
    procedure CheckBoxLockTimeClick(Sender: TObject);
    procedure TimerBlinkTimer(Sender: TObject);
  private
    FOnPlayingStarted     : TNotifyEvent;
    FOnPlayingStopped     : TNotifyEvent;
    FOnRecordingStarted   : TNotifyEvent;
    FOnRecordingStopped   : TNotifyEvent;
    FOnPauseStarted       : TNotifyEvent;
    FOnPauseStopped       : TNotifyEvent;
    FOnLoopingStarted     : TNotifyEvent;
    FOnLoopingStopped     : TNotifyEvent;
  private
    FShown                : Boolean;
    FColWidths            : array[ 0 .. 2] of Integer;
    FUseSplines           : Boolean;
  private
    FEditor               : TKnobsWirePanel;
    FAutomationData       : TKnobsEditHistory;
    FTimeIndex            : TKnobsTimeIndex;
    FEditDataTimeIndex    : TKnobsEditDataTimeIndex;
    FModuleNames          : TArray<string>;
    FControls             : TArray<TKnobsValuedControl>;
    FLockTime             : Boolean;
    FLockValue            : Boolean;
    FTime                 : TKnobsTimeStamp;
    FModuleName           : string;
    FZoomFactor           : Double;
    FCursorPosition       : TPoint;
    FRecording            : Boolean;
    FPlaying              : Boolean;
    FLooping              : Boolean;
    FReversed             : Boolean;
    FPaused               : Boolean;
    FAutomationChanged    : Boolean;
    FForceGridUpdate      : Boolean;
    FUsedWidth            : Integer;
    FUsedHeight           : Integer;
    FAssociate            : TKnobsValuedControl;
    FPlayPointer          : Integer;
  private
    FLastTickTime         : TKnobsTimeStamp;
    FTickTimeDelta        : TKnobsTimeStamp;
    FlargestTickTimeDelta : TKnobsTimeStamp;
  private
    function    GetBottomHeight   : Integer;
    procedure   SetBottomHeight   ( aValue: Integer);
    function    GetDetailsWidth   : Integer;
    procedure   SetDetailsWidth   ( aValue: Integer);
    procedure   SetCursorPosition ( const aValue: TPoint);
    procedure   SetZoomFactor     ( aValue: Double );
    function    GetAutomationSize : Integer;
    procedure   SetAutomationSize ( aValue: Integer);
    procedure   FixButtonEnables;
    procedure   SetRecording      ( aValue: Boolean);
    procedure   SetPlaying        ( aValue: Boolean);
    procedure   SetLooping        ( aValue: Boolean);
    procedure   SetPaused         ( aValue: Boolean);
    function    GetFreeCount      : Integer;
    function    GetFillCount      : Integer;
    function    GetSize           : Integer;
    procedure   SetPlayPointer    ( aValue: Integer);
    function    GetUseAlternateTitles: Boolean;
    procedure   SetUseAlternateTitles( aValue: Boolean);
    procedure   SetUseSplines     ( aValue: Boolean);
    procedure   FixTimeNameLabel;
    procedure   SetTime           ( const aValue: TKnobsTimeStamp);
    procedure   SetModuleName     ( const aValue: string);
    procedure   SetAssociate      ( const aValue: TKnobsValuedControl);
    procedure   SetLockTime       ( aValue: Boolean);
    procedure   SetLockValue      ( aValue: Boolean);
  private
    procedure   FixTimes;
    procedure   FixGrid( const aPath: string);
    procedure   ListViewClicked( const anItem: TListItem);
    procedure   SplitItem( anItemIndex: Integer; var Col0, Col1, Col2: string);
    procedure   FixListView;
    procedure   FixGridFromData;
    procedure   ScrollCursorIntoView;
    procedure   ScrollPointInView( aPoint: TPoint);                                                            overload;
    procedure   ScrollPointInView( anX, anY: LongInt);                                                         overload;
    procedure   AddEditData( const aSender: TKnobsValuedControl; const aData: PKnobsEditData);
    procedure   StatusUpdate;
    procedure   BlinkUpdate;
    procedure   LessZoom;
    procedure   MoreZoom;
    procedure   HandleGridPointChanged( const aSender: TObject; const aPath: string; anIndex: Integer; anX, anY: Double);
  public
    procedure   Initialize( anEditor: TKnobsWirePanel);
    procedure   UpdateModuleNames( const anOldNames, aNewNames: TStringArray);
    procedure   ClearAutomation;
    procedure   ResetAutomation;
    procedure   TakeSnapShot;
    procedure   SendValues;
    procedure   DumpAutomation( const aFilename: string);
    procedure   Stop;
    procedure   GoToLeftStop;
    procedure   GoToRightStop;
    procedure   GoToUpStop;
    procedure   GoToDownStop;
    procedure   GoLeft;
    procedure   GoRight;
    procedure   GoUp;
    procedure   GoDown;
    procedure   RegisterCallbacks(
      anOnPlayingStarted   : TNotifyEvent;
      anOnPlayingStopped   : TNotifyEvent;
      anOnRecordingStarted : TNotifyEvent;
      anOnRecordingStopped : TNotifyEvent;
      anOnPauseStarted     : TNotifyEvent;
      anOnPauseStopped     : TNotifyEvent;
      anOnLoopingStarted   : TNotifyEvent;
      anOnLoopingStopped   : TNotifyEvent
    );
    procedure   UnregisterCallbacks;
    procedure   SaveIni( const aSectionName: string; const anIniFile: TMemIniFile);
    procedure   LoadIni( const aSectionName: string; const anIniFile: TMemIniFile);
    procedure   FillStatusFeedback( const aLabel: TLabel);
    procedure   FillLedFeedback( const aLed: TKnobsLed);
    procedure   FlagAutomationChanged;
    procedure   FixTransportButtons( aRecordButton, aPlayButton, aPauseButton, aLoopButton: TSpeedButton);
    procedure   SaveAutomation;
    procedure   LoadAutomation;
    procedure   FixAutomation;
    procedure   Tick( CurrentTime: TKnobsTimeStamp);
    procedure   Clock;
  private
    property    BottomHeight       : Integer             read GetBottomHeight       write SetBottomHeight;
    property    DetailsWidth       : Integer             read GetDetailsWidth       write SetDetailsWidth;
    property    Associate          : TKnobsValuedControl read FAssociate            write SetAssociate;
    property    UseAlternateTitles : Boolean             read GetUseAlternateTitles write SetUseAlternateTitles;
    property    UseSplines         : Boolean             read FUseSplines           write SetUseSplines;
    property    LockTime           : Boolean             read FLockTime             write SetLockTime;
    property    LockValue          : Boolean             read FLockValue            write SetLockValue;
  public
    property    AutomationSize     : Integer             read GetAutomationSize     write SetAutomationSize;
    property    Recording          : Boolean             read FRecording            write SetRecording;
    property    Playing            : Boolean             read FPlaying              write SetPlaying;
    property    Looping            : Boolean             read FLooping              write SetLooping;
    property    Reversed           : Boolean             read FReversed             write FReversed;
    property    Paused             : Boolean             read FPaused               write SetPaused;
    property    ZoomFactor         : Double              read FZoomFactor           write SetZoomFactor;
    property    CursorPosition     : TPoint              read FCursorPosition       write SetCursorPosition;
    property    Time               : TKnobsTimeStamp     read FTime                 write SetTime;
    property    ModuleName         : string              read FModuleName           write SetModuleName;
    property    FreeCount          : Integer             read GetFreeCount;
    property    FillCount          : Integer             read GetFillCount;
    property    PlayPointer        : Integer             read FPlayPointer          write SetPlayPointer;
    property    Size               : Integer             read GetSize;
  end;


var

  FormAutomation: TFormAutomation;


implementation


// User area

const

  X_GROWTH  = 32;
  Y_GROWTH  = 16;
  CELL_SIZE = 15;
  DIGITS    = [ '0' .. '9'];


{ ========
  TScrollBox = class( Vcl.Forms.TScrollBox)
  protected
}
    procedure   TScrollBox.CreateParams( var Params: TCreateParams); // override;
    begin
      inherited;
      Params.ExStyle := Params.ExStyle or WS_EX_LEFTSCROLLBAR;
    end;


{ ========
  TKnobsTimeIndexList = record
  private
    FTime     : TKnobsTimeStamp;
    FDuration : TKnobsTimeStamp;
    FItems    : TArray<Integer>;
  public
}

    procedure   TKnobsTimeIndexList.Create( aTime: TKnobsTimeStamp);
    begin
      FTime := aTime;
      Clear;
    end;


    procedure   TKnobsTimeIndexList.Clear;
    begin
      SetLength( FItems, 0);
    end;


    procedure   TKnobsTimeIndexList.AddItem( anItem: Integer);
    begin
      SetLength( FItems, Count + 1);
      FItems[ Count - 1] := anItem;
    end;


    function    TKnobsTimeIndexList.Count: Integer;
    begin
      Result := Length( FItems);
    end;


    procedure   TKnobsTimeIndexList.SetDuration( aValue: TKnobsTimeStamp);
    begin
      FDuration := aValue;
    end;


    function    TKnobsTimeIndexList.Time : TKnobsTimeStamp;
    begin
      Result := FTime;
    end;


    function    TKnobsTimeIndexList.Duration: TKnobsTimeStamp;
    begin
      Result := FDuration;
    end;


    function    TKnobsTimeIndexList.Item( anIndex: Integer): Integer;
    begin
      Result := FItems[ anIndex];
    end;


{ ========
  TTimeIndexRecords = TArray<TTimeIndexRec>;


  TTimeIndex = class
  private
    FData : TKnobsTimeIndexLists;
  public
    property    Count                        : Integer            read GetCount;
    property    StartTime                    : TKnobsTimeStamp    read GetStartTime;
    property    TimeSpan                     : TKnobsTimeStamp    read GetTimeSpan;
    property    Time      [ anIndex: Integer]: TKnobsTimeStamp    read GetTime;
    property    Duration  [ anIndex: Integer]: TKnobsTimeStamp    read GetDuration;
    property    ItemCount [ anIndex: Integer]: Integer            read GetItemCount;
    property    Item      [ i, j   : Integer]: Integer            read GetItem;
  private
}

    function    TKnobsTimeIndex.GetCount: Integer;
    begin
      Result := Length( FData);
    end;


    function    TKnobsTimeIndex.GetStartTime : TKnobsTimeStamp;
    begin
      if Count > 0
      then Result := Time[ 0]
      else Result := 0;
    end;


    function    TKnobsTimeIndex.GetTimeSpan: TKnobsTimeStamp;
    begin
      if Count > 0
      then Result := Time[ Count - 1] + Duration[ Count - 1] - Time[ 0]
      else Result := 0;
    end;


    function    TKnobsTimeIndex.GetTime( anIndex: Integer) : TKnobsTimeStamp;
    begin
      Result := FData[ anIndex].FTime;
    end;


    function    TKnobsTimeIndex.GetDuration( anIndex: Integer) : TKnobsTimeStamp;
    begin
      Result := FData[ anIndex].FDuration;
    end;


    function    TKnobsTimeIndex.GetItemCount( anIndex: Integer) : Integer;
    begin
      Result := FData[ anIndex].Count;
    end;


    function    TKnobsTimeIndex.GetItem( i, j: Integer): Integer;
    begin
      Result := FData[ i].Item( j);
    end;


    procedure   TKnobsTimeIndex.Sort;
    var
      Comparer : IComparer<TKnobsTimeIndexList>;
    begin
      Comparer :=
        TDelegatedComparer<TKnobsTimeIndexList>.Create
        (
          function( const aLeft, aRight: TKnobsTimeIndexList): Integer
          begin
            Result := aLeft.Time - aRight.Time;
          end
        );

      TArray.Sort<TKnobsTimeIndexList>( FData, Comparer);
    end;


//  public

    procedure   TKnobsTimeIndex.AddItem( const aTimeIndexList: TKnobsTimeIndexList); // overload;
    begin
      SetLength( FData, Count + 1);
      FData[ Count - 1] := aTimeIndexList;
    end;


    procedure   TKnobsTimeIndex.AddItem( aTime: TKnobsTimeStamp; anIndex: Integer); // overload;
    var
      T : TKnobsTimeIndexList;
    begin
      T.Create( aTime);
      T.AddItem( anIndex);
      AddItem( T);
    end;


    procedure   TKnobsTimeIndex.Clear;
    var
      i : Integer;
    begin
      for i := 0 to Length( FData) - 1
      do FData[ i].Clear;

      SetLength( FData, 0);
    end;


    procedure   TKnobsTimeIndex.SetFromHistory( const aData: TKnobsEditHistory);
    var
      i         : Integer;
      p         : Integer;
      aLasttime : TKnobsTimeStamp;
    begin
      Clear;

      if  Assigned( aData)
      and ( aData.FillCount > 0)
      then begin
        for i := 1 to aData.FillCount - 1
        do begin
          p := FindTime( aData[ i].GetTimeStamp);

          if p >= 0
          then FData[ p].AddItem( i)
          else AddItem( aData[ i].GetTimeStamp, i)
        end;

        Sort;

        if Length( FData) > 0
        then begin
          aLastTime := FData[ 0].Time;

          for i := 1 to Length( FData) - 1
          do begin
            FData[ i - 1].SetDuration( FData[ i].Time - aLasttime);
            aLastTime := FData[ i].Time;
          end;

          FData[ Length( FData) - 1].SetDuration( 0);
        end;
      end;
    end;


    function    TKnobsTimeIndex.FindTime( aTime: TKnobsTimeStamp): Integer;
    var
      i : Integer;
    begin
      Result := -1;

      for i := 0 to Count - 1
      do begin
        if aTime = Time[ i]
        then begin
          Result := i;
          Break;
        end;
      end;
    end;


{ ========
  TKnobsEditDataTimeList = class
  private
    FItems : TArray<PKnobsEditData>;
  public
    property    Count                   : Integer         read GetCount;
    property    Time[ anIndex: Integer] : TKnobsTimeStamp read GetTime;
    property    Item[ anIndex: Integer] : PKnobsEditData  read GetItem;
  private
}

    function    TKnobsEditDataTimeList.GetCount: Integer;
    begin
      Result := Length( FItems);
    end;


    function    TKnobsEditDataTimeList.GetTime( anIndex: Integer): TKnobsTimeStamp;
    begin
      Result := FItems[ anIndex]^.GetTimeStamp;
    end;


    function    TKnobsEditDataTimeList.GetItem( anIndex: Integer): PKnobsEditData;
    begin
      Result := FItems[ anIndex];
    end;


//  public

    constructor TKnobsEditDataTimeList.Create;
    begin
      inherited Create;
    end;


    destructor  TKnobsEditDataTimeList.Destroy; // override;
    begin
      SetLength( FItems, 0);
      inherited;
    end;


    procedure   TKnobsEditDataTimeList.AddItem( anItem: PKnobsEditData);
    var
      aTime : TKnobsTimeStamp;
      p     : Integer;
    begin
      if Assigned( anItem)
      then begin
        aTime := anItem^.GetTimeStamp;
        p     := FindTime( aTime);

        if p < 0
        then begin
          SetLength( FItems, Count + 1);
          p := Count - 1;
        end;

        FItems[ p] := anItem;
      end;
    end;


    function    TKnobsEditDataTimeList.FindTime( aTime: TKnobsTimeStamp): Integer;
    var
      i : Integer;
    begin
      Result := -1;

      for i := 0 to Count - 1
      do begin
        if Time[ i] = aTime
        then begin
          Result := i;
          Break;
        end;
      end;
    end;


    procedure   TKnobsEditDataTimeList.FillGrid( const aGrid: TKnobsDataMaker; aStartTime, aTimeSpan: TKnobsTimeStamp);
    // todo : time scaling s not right thos way
    var
      i     : Integer;
      aData : TSignalPairArray;
    begin
      if  Assigned( aGrid)
      and ( aTimeSpan > 0)
      then begin
        SetLength( aData, Count);

        for i := 0 to Count - 1
        do begin
          if  Assigned( FItems[ i])
          and ( FItems[ i]^.GetSteps > 0)
          then begin
            aData[ i].Y := FItems[ i]^.GetPosition / FItems[ i]^.GetSteps;
            aData[ i].X := ( FItems[ i]^.GetTimeStamp - aStartTime) / aTimeSpan;
          end;
        end;

        aGrid.Data := aData;
      end;
    end;


{ ========
  TKnobsEditDataTimeIndex = class
  private
    FData      : TArray<TKnobsEditDataTimeList>;
    FPathIndex : TStringList;
  public
    property    Count                                       : Integer                read GetCount;
    property    Data     [ anIndex: Integer               ] : TKnobsEditDataTimeList read GetData;
    property    TimeCount[ anIndex: Integer               ] : Integer                read GetTimeCount;
    property    Time     [ aDataIndex, aTimeIndex: Integer] : TKnobsTimeStamp        read GetTime;
  private
}

    function    TKnobsEditDataTimeIndex.GetCount: Integer;
    begin
      Result := Length( FData);
    end;


    function    TKnobsEditDataTimeIndex.GetData( anIndex: Integer): TKnobsEditDataTimeList;
    begin
      Result := FData[ anIndex];
    end;


    function    TKnobsEditDataTimeIndex.GetTimeCount( anIndex: Integer): Integer;
    begin
      Result := FData[ anIndex].Count;
    end;


    function    TKnobsEditDataTimeIndex.GetTime( aDataIndex, aTimeIndex: Integer): TKnobsTimeStamp;
    begin
      Result := FData[ aDataIndex].Time[ aTimeIndex];
    end;


//  public

    constructor TKnobsEditDataTimeIndex.Create;
    begin
      inherited;
      FPathIndex := TStringList.Create;
      FPathIndex.Duplicates    := dupError;
      FPathIndex.Sorted        := True;
      FPathIndex.CaseSensitive := False;
      Clear;
    end;


    destructor  TKnobsEditDataTimeIndex.Destroy; // override;
    begin
      Clear;
      FreeAndNil( FPathIndex);
      inherited;
    end;


    procedure   TKnobsEditDataTimeIndex.Clear;
    var
      i : Integer;
    begin
      for i := 0 to Count - 1
      do FreeAndNil( FData[ i]);

      SetLength( FData, 0);
      FPathIndex.Clear;
    end;


    procedure   TKnobsEditDataTimeIndex.SetFromHistory( const aData: TKnobsEditHistory);
    var
      i : Integer;
      p : Integer;
    begin
      Clear;

      if Assigned( aData)
      then begin
        for i := 0 to aData.FillCount - 1
        do begin
          p := FindPath( aData[ i].GetPath);

          if p < 0
          then begin
            p := Count;
            SetLength( FData, Count + 1);
            FData[ p] := TKnobsEditDataTimeList.Create;
            FPathIndex.AddObject( aData[ i].GetPath, Pointer( p));
          end;

          FData[ p].AddItem( aData.Data[ i]);
        end;
      end;
    end;


    function    TKnobsEditDataTimeIndex.FindPath( const aPath: string): Integer;
    begin
      Result := FPathIndex.IndexOf( aPath);

      if Result >= 0
      then Result := Integer( FPathIndex.Objects[ Result]);
    end;


    procedure   TKnobsEditDataTimeIndex.FillGrid( anIndex: Integer; const aGrid: TKnobsDataMaker; aStartTime, aTimeSpan: TKnobsTimeStamp);
    begin
      FData[ anIndex].FillGrid( aGrid, aStartTime, aTimeSpan);
    end;


{ ========
  TFormAutomation = class(TForm)
  private
}

    function    TFormAutomation.GetBottomHeight: Integer;
    begin
      Result := PanelBottom.Height;
    end;


    procedure   TFormAutomation.SetBottomHeight( aValue: Integer);
    begin
      PanelBottom.Height := aValue;
    end;


    function    TFormAutomation.GetDetailsWidth: Integer;
    begin
      Result := PanelDetails.Width;
    end;


    procedure   TFormAutomation.SetDetailsWidth( aValue: Integer);
    begin
      PanelDetails.Width := aValue;
    end;


    procedure   TFormAutomation.SetCursorPosition( const aValue: TPoint);
    var
      P : TPoint;
    begin
      if KnobsAutomationGrid.DataWidth > 0
      then begin
        P.X := Clip( aValue.X, -1, FUsedWidth );
        P.Y := Clip( aValue.Y, -1, FUsedHeight);
      end
      else begin
        P.X := -1;
        P.Y := -1;
      end;

      if ( P.X <> FCursorPosition.X)
      or ( P.Y <> FCursorPosition.Y)
      then begin
        if  ( P.X >= 0)
        and ( P.X < FTimeIndex.Count)
        then Time := FTimeIndex.Time[ P.X]
        else Time := -1;

        if  ( P.Y >= 0)
        and ( P.Y < Length( FModuleNames))
        then ModuleName := FModuleNames[ P.Y]
        else ModuleName := '-';

        FCursorPosition.X := P.X;
        FCursorPosition.Y := P.Y;
        KnobsAutomationGrid.CursorX := P.X / KnobsAutomationGrid.DataWidth;
        KnobsAutomationGrid.CursorY := P.Y / KnobsAutomationGrid.DataHeight;
        ScrollCursorIntoView;

        if not Playing
        then FixListView;

        FixTimeNameLabel;
        FillStatusFeedback( LabelAutomationInfo)
      end;
    end;


    procedure   TFormAutomation.SetZoomFactor( aValue: Double );
    begin
      if aValue <> FZoomFactor
      then begin
        FZoomFactor        := aValue;
        LabelZoom.Caption  := Format( 'zoom %.3g', [ ZoomFactor], AppLocale);
        FForceGridUpdate   := True;
        FAutomationChanged := True;
      end;
    end;


    function    TFormAutomation.GetAutomationSize: Integer;
    begin
      if Assigned( FAutomationData)
      then Result := FAutomationData.Size
      else Result := 0;
    end;


    procedure   TFormAutomation.SetAutomationSize( aValue: Integer);
    begin
      if  Assigned( FAutomationData)
      and ( aValue <> AutomationSize)
      then begin
        Recording               := False;
        Playing                 := False;
        FAutomationData.Size    := aValue;
        EditAutomationSize.Text := IntToStr( AutomationSize);
        ClearAutomation;
      end;
    end;


    procedure   TFormAutomation.FixButtonEnables;
    begin
      FixTransportButtons( SpeedButtonRecord, SpeedButtonPlay, SpeedButtonPause, SpeedButtonLoop);
    end;


    procedure   TFormAutomation.SetRecording( aValue: Boolean);
    begin
      if aValue <> FRecording
      then begin
        FRecording             := aValue;
        SpeedButtonRecord.Down := FRecording;

        if FRecording
        then begin
          Playing := False;

          if Assigned( FOnRecordingStarted)
          then FOnRecordingStarted( Self);
        end
        else begin
          if Assigned( FOnRecordingStopped)
          then FOnRecordingStopped( Self);
        end;

        if  Assigned( FEditor)
        and FRecording
        then begin
          if  Assigned( FAutomationData)
          and ( FAutomationData.FillCount = 0)
          then FEditor.CreateValueSnapShot( FAutomationData);

          FAutomationChanged := True;
        end;

        FixButtonEnables;
      end
    end;


    procedure   TFormAutomation.SetPlaying( aValue: Boolean);
    begin
      if aValue <> FPlaying
      then begin
        FPlaying             := aValue;
        SpeedButtonPlay.Down := FPlaying;

        if FPlaying
        then begin
          Recording := False;
          FixTimes;

          if Assigned( FOnPlayingStarted)
          then FOnPlayingStarted( Self);
        end
        else begin
          if Assigned( FOnPlayingStopped)
          then FOnPlayingStopped( Self);

          FixListView;

          if Reversed
          then PlayPointer := FUsedWidth
          else PlayPointer := -1;
        end;

        FixButtonEnables;
      end
    end;


    procedure   TFormAutomation.SetLooping ( aValue: Boolean);
    begin
      if aValue <> FLooping
      then begin
        FLooping := aValue;
        SpeedButtonLoop.Down := FLooping;

        if FLooping
        then begin
          if Assigned( FOnLoopingStarted)
          then FOnLoopingStarted( Self);
        end
        else begin
          if Assigned( FOnLoopingStopped)
          then FOnLoopingStopped( Self);
        end;

        FixButtonEnables;
      end;
    end;


    procedure   TFormAutomation.SetPaused( aValue: Boolean);
    begin
      if aValue <> FPaused
      then begin
        FPaused := aValue;
        SpeedButtonPause.Down := FPaused;

        if FPaused
        then begin
          if Assigned( FOnPauseStarted)
          then FOnPauseStarted( Self);
        end
        else begin
          if Assigned( FOnPauseStopped)
          then FOnPauseStopped( Self);
        end;

        FixButtonEnables;
      end;
    end;


    function    TFormAutomation.GetFreeCount: Integer;
    begin
      if Assigned( FAutomationData)
      then Result := FAutomationData.FreeCount
      else Result := 0;
    end;


    function    TFormAutomation.GetFillCount: Integer;
    begin
      if Assigned( FAutomationData)
      then Result := FAutomationData.FillCount
      else Result := 0;
    end;


    function    TFormAutomation.GetSize: Integer;
    begin
      if Assigned( FAutomationData)
      then Result := FAutomationData.Size
      else Result := 0;
    end;


    procedure   TFormAutomation.SetPlayPointer( aValue: Integer);
    begin
      if aValue <> FPlayPointer
      then begin
        FPlayPointer   := aValue;
        CursorPosition := Point( aValue, CursorPosition.Y);
      end;
    end;


    function    TFormAutomation.GetUseAlternateTitles: Boolean;
    begin
      if Assigned( FEditor)
      then Result := FEditor.UseAlternateTitles
      else Result := False;
    end;


    procedure   TFormAutomation.SetUseAlternateTitles( aValue: Boolean);
    begin
      if Assigned( FEditor)
      then FEditor.UseAlternateTitles := aValue;
    end;


    procedure   TFormAutomation.SetUseSplines( aValue: Boolean);
    begin
      if aValue <> FUseSplines
      then begin
        FUseSplines := aValue;

        if Assigned( KnobsDataMaker)
        then KnobsDataMaker.UseSplines := aValue;

        if Assigned( datagraph_splines)
        then datagraph_splines.KnobPosition := Ord( aValue);
      end;
    end;


    procedure   TFormAutomation.FixTimeNameLabel;
    begin
      LabelTime.Caption := Format( '%s %s', [ FormatTime64ms( FTime), ModuleName], AppLocale);
    end;


    procedure   TFormAutomation.SetTime( const aValue: TKnobsTimeStamp);
    begin
      if aValue <> FTime
      then begin
        FTime := aValue;
        FixTimeNameLabel;
      end;
    end;


    procedure   TFormAutomation.SetModuleName( const aValue: string);
    begin
      if aValue <> FModuleName
      then begin
        FModuleName := aValue;
        FixTimeNameLabel;
      end;
    end;


    procedure   TFormAutomation.SetAssociate( const aValue: TKnobsValuedControl);
    begin
      if aValue <> FAssociate
      then begin
        FAssociate := aValue;
      end;
    end;



    procedure   TFormAutomation.SetLockTime( aValue: Boolean);
    begin
      if aValue <> FLockTime
      then begin
        FLockTime                    :=     aValue;
        CheckBoxLockTime.Checked     :=     aValue;
        KnobsDataMaker  .AllowXMoves := not aValue;

        if aValue
        then LockValue := False;
      end;
    end;


    procedure    TFormAutomation.SetLockValue( aValue: Boolean);
    begin
      if aValue <> FLockValue
      then begin
        FLockValue                    :=     aValue;
        CheckBoxLockValue.Checked     :=     aValue;
        KnobsDataMaker   .AllowYMoves := not aValue;

        if aValue
        then LockTime := False;
      end;
    end;


// private

    procedure   TFormAutomation.FixTimes;
    begin
      if Assigned( FAutomationData)
      then begin
        FAutomationData.FixTimes;
        FixGridFromData;
      end;
    end;


    procedure   TFormAutomation.FixGrid( const aPath: string);
    var
      p     : Integer;
      aData : TSignalPairArray;
    begin
      if  Assigned( FEditDataTimeIndex)
      and Assigned( FTimeIndex        )
      then begin
        p := FEditDataTimeIndex.FindPath( aPath);

        if p >= 0
        then FEditDataTimeIndex.FillGrid( p, KnobsDataMaker, FTimeIndex.StartTime, FTimeIndex.TimeSpan)
        else begin
          SetLength( aData, 0);
          KnobsDataMaker.Data := aData;
        end;
      end;
    end;


    procedure   TFormAutomation.ListViewClicked( const anItem: TListItem);
    var
      p     : Integer;
      aPath : string;
    begin
      Associate := nil;

      if  Assigned( anItem)
      then begin
        p := Integer( anItem.Data);

        if  ( p >= 0)
        and ( p < FAutomationData.FillCount)
        and ( anItem.SubItems.Count >= 2)
        then begin
          Associate := FAutomationData[ p].GetAssociate;

          if Assigned( Associate)
          then begin
            if Assigned( FEditor)
            then FEditor.SetActiveInDesigner( Format( '%s_%s', [ Associate.ModuleName, Associate.Name], AppLocale), True);
          end;

          aPath := FAutomationData[ p].GetPath;
          FixGrid( aPath);
        end;
      end
      else begin
        if Assigned( FEditor)
        then FEditor.SetActiveInDesigner( '', True);

        FixGrid( '');
      end;
    end;


    procedure   TFormAutomation.SplitItem( anItemIndex: Integer; var Col0, Col1, Col2: string);
    var
      aParts : TStringList;
      aData  : TKnobsEditData;
    begin
      Col0 := '?';
      Col1 := '';
      Col2 := '';

      try
        if  ( anItemIndex >= 0)
        and ( anItemIndex < FAutomationData.Size)
        then begin
          aParts := Explode( FAutomationData[ anItemIndex].GetPath, ' ');

          try
            TrimStrings( aParts);

            if aParts.Count = 4
            then begin
              Col0 := Format( '%s %s %s', [ aParts[ 0], aParts[ 1], aParts[ 2]], AppLocale);
              Col1 := aParts[ 3];
            end;
          finally
            aParts.DisposeOf;
          end;

          aData := FAutomationData[ anItemIndex];
          Col2  := Converters.PosToDisplay( aData.GetControlType, aData.GetPosition, aData.GetSteps);
        end;
      except
        on E: Exception
        do col2 := E.ToString;
      end;
    end;


    procedure   TFormAutomation.FixListView;
    var
      i    : Integer;
      p    : Integer;
      Item : TListItem;
      Col0 : string;
      Col1 : string;
      Col2 : string;
      M    : string;
    begin
      ListView.Clear;
      Associate := nil;

      if  Assigned( FAutomationData)
      and ( Time >= 0)
      then begin
        p := -1;
        M := Format( 'mod %d ', [ CursorPosition.Y], AppLocale);

        for i := 0 to FAutomationData.FillCount - 1
        do begin
          if FAutomationData[ i].GetTimeStamp = Time
          then begin
            p := i;
            Break;
          end;
        end;

        if p >= 0
        then begin
          while ( p < FAutomationData.FillCount)
          do begin
            SplitItem( p, Col0, Col1, Col2);

            if  Col0.Contains( M)
            and ( FAutomationData[ p].GetTimeStamp = Time)
            then begin
              Item := ListView.Items.Add;
              Item.Data       := Pointer( p);
              Item.Caption    := Col0;
              Item.SubItems.Add( Col1);
              Item.SubItems.Add( Col2);
            end;

            Inc( p);
          end;
        end;

        if ListView.Items.Count > 0
        then begin
          ListView.ItemIndex := 0;
          ListviewClicked( ListView.Items[ 0]);
        end;
      end;
    end;


    procedure   TFormAutomation.FixGridFromData;
    var
      i             : Integer;
      j             : Integer;
      p             : Integer;
      aData         : TKnobsEditData;
      aName         : string;
      aModule       : TKnobsCustomModule;
      aControl      : TControl;
      Y             : Integer;
      aControlIndex : Integer;
      aCursor       : TPoint;
    begin
      if  Assigned( FAutomationData   )
      and Assigned( FEditor           )
      and Assigned( FTimeIndex        )
      and Assigned( FEditDataTimeIndex)
      then begin

        // Calculate amount of columns needed, rowcount is just the anount of modules in the patch

        aControlIndex := 0;
        FUsedHeight   := FEditor.ModuleCount;
        SetLength( FModuleNames, FUsedHeight);
        FTimeIndex        .SetFromHistory( FAutomationData);
        FEditDataTimeIndex.SetFromHistory( FAutomationData);
        FUsedWidth := FTimeIndex.Count;

        for i := 0 to FUsedHeight - 1
        do begin
          aModule := FEditor.Module[ i];

          for j := 0 to aModule.ControlCount - 1
          do begin
            aControl := aModule.Controls[ j];

            if  ( aControl is TKnobsValuedControl)
            and FAutomationData.CanRecord( TKnobsValuedControl( aControl))
            then Inc( aControlIndex);
          end;
        end;

        SetLength( FControls, aControlIndex);
        aControlIndex := 0;

        for i := 0 to FUsedHeight - 1
        do begin
          aModule := FEditor.Module[ i];
          FModuleNames[ i] := Format( '%s  %s', [ aModule.FriendlyName, aModule.Title], AppLocale);

          for j := 0 to aModule.ControlCount - 1
          do begin
            aControl := aModule.Controls[ j];

            if  ( aControl is TKnobsValuedControl)
            and FAutomationData.CanRecord( TKnobsValuedControl( aControl))
            then begin
              FControls[ aControlIndex] := TKnobsValuedControl( aControl);
              Inc( aControlIndex);
            end;
          end;
        end;

        p := 0;

        while p <= FEditor.ModuleCount
        do begin
          p := p + Y_GROWTH;
          FForceGridUpdate := True;
        end;

        KnobsAutomationGrid.DataHeight := p;
        p := 0;

        while p <= FUsedWidth
        do begin
          p := p + X_GROWTH;
          FForceGridUpdate := True;
        end;

        KnobsAutomationGrid.DataWidth := p;

        if FForceGridUpdate
        then begin
          KnobsAutomationGrid.Height := Ceil( CELL_SIZE * FZoomFactor * KnobsAutomationGrid.DataHeight);
          KnobsAutomationGrid.Width  := Ceil( CELL_SIZE * FZoomFactor * KnobsAutomationGrid.DataWidth);
        end;

        KnobsAutomationGrid.ClearData;

        for i := 0 to FTimeIndex.Count - 1
        do begin
          for j := 0 to FTimeIndex.ItemCount[ i] - 1
          do begin
            aData := FAutomationData[ FTimeIndex.Item[ i, j]];
            aName := aData.GetPath;

            p := 1;
            Y := 0;

            while ( p <= Length( aName))
            and   ( aName[ p] <> ' ')
            do    Inc( p);

            while ( p <= Length( aName))
            and   ( aName[ p] = ' ')
            do    Inc( p);

            while ( p <= Length( aName))
            and   ( CharInSet( aName[ p], DIGITS))
            do begin
              Y := 10 * Y + (Ord( aName[ p]) - Ord( '0'));
              Inc( p);
            end;

            KnobsAutomationGrid.Pixel[ i, Y] := True;
          end;
        end;

        FForceGridUpdate := False;
      end;

      if not Playing
      then begin
        FixListView;
        aCursor := CursorPosition;

        aCursor.X := Clip( aCursor.X, -1, FUsedWidth );
        aCursor.Y := Clip( aCursor.Y, -1, FUsedHeight);
        CursorPosition := aCursor;
      end;
    end;


    procedure   TFormAutomation.ScrollCursorIntoView;
    var
      P : TPoint;
    begin
      P.X := Round(( 0.5 + CursorPosition.X) * CELL_SIZE * ZoomFactor);
      P.Y := Round(( 0.5 + CursorPosition.Y) * CELL_SIZE * ZoomFactor);
      ScrollPointInView( ScrollBoxAutomationGrid.ScreenToClient( KnobsAutomationGrid.ClientToScreen( P)));
    end;


    procedure   TFormAutomation.ScrollPointInView( aPoint: TPoint); // overload;
    begin
      ScrollPointInView( aPoint.X, aPoint.Y);
    end;


    procedure   TFormAutomation.ScrollPointInView( anX, anY: LongInt); // overload;
    var
      aRect      : TRect;
      aClearance : Integer;
    begin
      aClearance := Ceil( CELL_SIZE * ZoomFactor);

      with ScrollBoxAutomationGrid
      do begin
        aRect.TopLeft     := Point( anX - aClearance, anY - aClearance);
        aRect.BottomRight := Point( anX + aClearance, anY + aClearance);
        Dec( aRect.Left  , HorzScrollBar.Margin);
        Inc( aRect.Right , HorzScrollBar.Margin);
        Dec( aRect.Top   , VertScrollBar.Margin);
        Inc( aRect.Bottom, VertScrollBar.Margin);

        if aRect.Left < 0
        then begin
          with HorzScrollBar
          do Position := Position + aRect.Left
        end
        else if aRect.Right > ClientWidth
        then begin
          if aRect.Right - aRect.Left > ClientWidth
          then aRect.Right := aRect.Left + ClientWidth;

          with HorzScrollBar
          do Position := Position + aRect.Right - ClientWidth;
        end;

        if aRect.Top < 0
        then begin
          with VertScrollBar
          do Position := Position + aRect.Top;
        end
        else if aRect.Bottom > ClientHeight
        then begin
          if aRect.Bottom - aRect.Top > ClientHeight
          then aRect.Bottom := aRect.Top + ClientHeight;

          with VertScrollBar
          do Position := Position + aRect.Bottom - ClientHeight;
        end;
      end;
    end;


    procedure   TFormAutomation.AddEditData( const aSender: TKnobsValuedControl; const aData: PKnobsEditData);
    begin
      if  Assigned( FAutomationData)
      and Recording
      and FAutomationData.CanRecord( aSender)
      and not Paused
      then begin
        FAutomationData.AddData( aData^);
        FAutomationChanged := True;
      end;
    end;


    procedure   TFormAutomation.StatusUpdate;
    begin
      if  Visible
      and FAutomationChanged
      then begin
        FAutomationChanged := False;
        FixGridFromData;
      end;
    end;


    procedure   TFormAutomation.BlinkUpdate;
    begin
      if Visible
      then begin
        FillLedFeedback( KnobsLedAutomationActive);

        LabelDebug.Caption :=
          Format(
            'largest time slip: %d ms, run time %s',
            [
              FlargestTickTimeDelta - 1,
              FormatTime64( FLastTickTime)
            ],
            AppLocale
          );
      end;
    end;


    procedure   TFormAutomation.LessZoom;
    begin
      if ZoomFactor > 0.5
      then ZoomFactor := ZoomFactor / Sqrt( 2);
    end;


    procedure   TFormAutomation.MoreZoom;
    begin
      if ZoomFactor < 4.01
      then ZoomFactor := ZoomFactor * Sqrt( 2);
    end;


    procedure   TFormAutomation.HandleGridPointChanged( const aSender: TObject; const aPath: string; anIndex: Integer; anX, anY: Double);
    var
      p           : Integer;
      aCurrPath   : string;
      aData       : PKnobsEditData;
      aNewData    : TKnobsEditData;
      anAssoc     : TKnobsValuedControl;
      aCtrlType   : string;
      aNewPos     : Integer;
      aNewTime    : TKnobsTimeStamp;
      TimeChanged : Boolean;
      anOldY      : TSignal;
    begin
      if  Assigned( FEditDataTimeIndex)
      and Assigned( Associate         )
      and Assigned( FTimeIndex        )
      and ( KnobsAutomationGrid.DataWidth > 0)
      then begin
        KnobsAutomationGrid.CursorX := anIndex;
        aCurrPath := Format( '%s    %s   %s', [ ModuleLongToFriendlyName( Associate.ModuleName), Associate.ModuleTitle, Associate.ShortName], AppLocale);
        p         := FEditDataTimeIndex.FindPath( aCurrPath);

        if p >= 0
        then begin
          aData := FEditDataTimeIndex.Data[ p].Item[ anIndex];

          if Assigned( aData)
          then begin
            anAssoc := aData^.GetAssociate;

            if Assigned( anAssoc)
            then begin
              aCtrlType := aData^.GetControlType;

              if LockValue
              then aNewPos := aData^.GetPosition
              else aNewPos := Round( anY * aData^.GetSteps);

              if LockTime
              then aNewTime := aData^.GetTimeStamp
              else aNewTime := Round( FTimeIndex.TimeSpan * anX) + FTimeIndex.StartTime;

              TimeChanged := aNewTime <> aData^.GetTimeStamp;
              anAssoc.KnobPosition := aNewPos;
              aNewData.Create( aData^.GetAssociate, aNewTime, aData^.GetKind, aData^.GetPath);
              FEditDataTimeIndex.Data[ p].Item[ anIndex]^ := aNewData;
              anOldY := KnobsAutomationGrid.CursorY;

              if TimeChanged
              then FixGridFromData;

              KnobsAutomationGrid.CursorY := anOldY;
              KnobsAutomationGrid.CursorX := FTimeIndex.FindTime( aNewTime) / KnobsAutomationGrid.DataWidth;
            end;
          end;
        end;
      end;
    end;


// public

    procedure   TFormAutomation.Initialize( anEditor: TKnobsWirePanel);
    begin
      if Assigned( FEditor)
      then FEditor.OnAddEditData := nil;

      FEditor := anEditor;

      if Assigned( FEditor)
      then FEditor.OnAddEditData := AddEditData;
    end;


    procedure   TFormAutomation.UpdateModuleNames( const anOldNames, aNewNames: TStringArray);
    begin
      if Assigned( FAutomationData)
      then begin
        FAutomationData.UpdateNames( anOldNames, aNewNames);
        FAutomationChanged := True;
      end;
    end;


    procedure   TFormAutomation.ClearAutomation;
    begin
      if  Assigned( FAutomationData)
      and not Playing
      then begin
        FAutomationData.Clear;

        if FRecording
        then TakeSnapShot
        else FAutomationChanged := True;
      end;
    end;


    procedure   TFormAutomation.ResetAutomation;
    begin
      if  Assigned( FAutomationData)
      and not Recording
      then PlayPointer := 0;
    end;


    procedure   TFormAutomation.TakeSnapShot;
    begin
      if  Assigned( FAutomationData)
      and Assigned( FEditor)
      and not Playing
      then begin
        FEditor.CreateValueSnapShot( FAutomationData);
        FixGridFromData;
        PlayPointer := FTimeIndex.Count - 1;
      end;
    end;


    procedure   TFormAutomation.SendValues;
    var
      i     : Integer;
      aData : TKnobsEditData;
    begin
      if  Assigned( FAutomationData)
      and Assigned( FEditor        )
      and Assigned( FTimeIndex     )
      and ( FTimeIndex.Count > 0   )
      then begin
        for i := 0 to FTimeIndex.ItemCount[ CursorPosition.X] - 1
        do begin
          aData := FAutomationData[ FTimeIndex.FData[ CursorPosition.X].FItems[ i]];
          aData.SetAssociateFromData;
        end;
      end;
    end;


    procedure   TFormAutomation.DumpAutomation( const aFilename: string);
    var
      aStringList : TStringList;
    begin
      if   Assigned( FEditor)
      and  Assigned( FAutomationData)
      then begin
        aStringList := TStringList.Create;

        try
          FAutomationData.FillStrings( aStringList, True, FAutomationData.FillCount);
          aStringList.SaveToFile( aFileName);
        finally
          aStringList.DisposeOf;
        end;
      end;
    end;


    procedure   TFormAutomation.Stop;
    begin
      Recording := False;
      Playing   := False;
      Paused    := False;
    end;


    procedure   TFormAutomation.GoToLeftStop;
    var
      P     : TPoint;
      State : Boolean;
    begin
      P     := CursorPosition;
      State := KnobsAutomationGrid.Pixel[ P.X, P.Y];

      while ( P.X > 0)
      and   ( State = KnobsAutomationGrid.Pixel[ P.X, P.Y])
      do    P.X := P.X - 1;

      while ( P.X > 0)
      and   ( State <> KnobsAutomationGrid.Pixel[ P.X, P.Y])
      do    P.X := P.X - 1;

      if ( State <> KnobsAutomationGrid.Pixel[ P.X, P.Y])
      then CursorPosition := P;
    end;


    procedure   TFormAutomation.GoToRightStop;
    var
      P     : TPoint;
      State : Boolean;
    begin
      P     := CursorPosition;
      State := KnobsAutomationGrid.Pixel[ P.X, P.Y];

      while ( P.X < FUsedWidth)
      and   ( KnobsAutomationGrid.Pixel[ P.X, P.Y] = State)
      do    P.X := P.X + 1;

      while ( P.X < FUsedWidth)
      and   ( State <> KnobsAutomationGrid.Pixel[ P.X, P.Y])
      do    P.X := P.X + 1;

      if ( State <> KnobsAutomationGrid.Pixel[ P.X, P.Y])
      then CursorPosition := P;
    end;


    procedure   TFormAutomation.GoToUpStop;
    var
      P     : TPoint;
      State : Boolean;
    begin
      P     := CursorPosition;
      State := KnobsAutomationGrid.Pixel[ P.X, P.Y];

      while ( P.Y > 0)
      and   ( State = KnobsAutomationGrid.Pixel[ P.X, P.Y])
      do    P.Y := P.Y - 1;

      if ( State <> KnobsAutomationGrid.Pixel[ P.X, P.Y])
      then CursorPosition := P;
    end;


    procedure   TFormAutomation.GoToDownStop;
    var
      P     : TPoint;
      State : Boolean;
    begin
      P     := CursorPosition;
      State := KnobsAutomationGrid.Pixel[ P.X, P.Y];

      while ( P.Y < FUsedHeight)
      and   ( State = KnobsAutomationGrid.Pixel[ P.X, P.Y])
      do    P.Y := P.Y + 1;

      if ( State <> KnobsAutomationGrid.Pixel[ P.X, P.Y])
      then CursorPosition := P;
    end;


    procedure   TFormAutomation.GoLeft;
    begin
      if CursorPosition.X > 0
      then CursorPosition := Point( CursorPosition.X - 1, CursorPosition.Y);
    end;


    procedure   TFormAutomation.GoRight;
    begin
      if CursorPosition.X < FUsedWidth
      then CursorPosition := Point( CursorPosition.X + 1, CursorPosition.Y);
    end;


    procedure   TFormAutomation.GoUp;
    begin
      if CursorPosition.Y > 0
      then CursorPosition := Point( CursorPosition.X, CursorPosition.Y - 1);
    end;


    procedure   TFormAutomation.GoDown;
    begin
      if CursorPosition.Y < FUsedHeight
      then CursorPosition := Point( CursorPosition.X, CursorPosition.Y + 1);
    end;


    procedure   TFormAutomation.RegisterCallbacks(
      anOnPlayingStarted   : TNotifyEvent;
      anOnPlayingStopped   : TNotifyEvent;
      anOnRecordingStarted : TNotifyEvent;
      anOnRecordingStopped : TNotifyEvent;
      anOnPauseStarted     : TNotifyEvent;
      anOnPauseStopped     : TNotifyEvent;
      anOnLoopingStarted   : TNotifyEvent;
      anOnLoopingStopped   : TNotifyEvent
    );
    begin
      FOnPlayingStarted   := anOnPlayingStarted  ;
      FOnPlayingStopped   := anOnPlayingStopped  ;
      FOnRecordingStarted := anOnRecordingStarted;
      FOnRecordingStopped := anOnRecordingStopped;
      FOnPauseStarted     := anOnPauseStarted    ;
      FOnPauseStopped     := anOnPauseStopped    ;
      FOnLoopingStarted   := anOnLoopingStarted  ;
      FOnLoopingStopped   := anOnLoopingStopped  ;
    end;


    procedure   TFormAutomation.UnregisterCallbacks;
    begin
      FOnPlayingStarted   := nil;
      FOnPlayingStopped   := nil;
      FOnRecordingStarted := nil;
      FOnRecordingStopped := nil;
      FOnPauseStarted     := nil;
      FOnPauseStopped     := nil;
      FOnLoopingStarted   := nil;
      FOnLoopingStopped   := nil;
    end;


    procedure   TFormAutomation.SaveIni( const aSectionName: string; const anIniFile: TMemIniFile);
    var
      i : Integer;
    begin
      if Assigned( anIniFile)
      then begin
        with anIniFile
        do begin
          EraseSection( aSectionName);
          WriteInteger( aSectionName, 'Left'          , Left                  );
          WriteInteger( aSectionName, 'Top'           , Top                   );
          WriteInteger( aSectionName, 'Width'         , Width                 );
          WriteInteger( aSectionName, 'Height'        , Height                );
          WriteInteger( aSectionName, 'WindowState'   , Integer( WindowState) );
          WriteFloat  ( aSectionName, 'ZoomFactor'    , ZoomFactor            );
          WriteInteger( aSectionName, 'BottomHeight'  , BottomHeight          );
          WriteInteger( aSectionName, 'DetailsWidth'  , DetailsWidth          );
          WriteInteger( aSectionName, 'AutomationSize', AutomationSize        );
          WriteBool   ( aSectionName, 'UseSplines'    , UseSplines            );
          WriteBool   ( aSectionName, 'LockTime'      , LockTime              );
          WriteBool   ( aSectionName, 'LockValue'     , LockValue             );

          if FShown
          then begin
            for i := 0 to 2
            do WriteInteger( aSectionName, Format( 'ColWidth%d', [ i], AppLocale), ListView.Columns[ i].Width);
          end
          else begin
            for i := 0 to 2
            do WriteInteger( aSectionName, Format( 'ColWidth%d', [ i], AppLocale), FColWidths[ i]);
          end;
        end;
      end;
    end;


    procedure   TFormAutomation.LoadIni( const aSectionName: string; const anIniFile: TMemIniFile);
    var
      L, T, W, H : Integer;
      WinState   : TWindowState;
      i          : Integer;
    begin
      if Assigned( anIniFile)
      then begin
        with anIniFile
        do begin
          L              :=                ReadInteger( aSectionName, 'Left'          , Left                  );
          T              :=                ReadInteger( aSectionName, 'Top'           , Top                   );
          W              :=                ReadInteger( aSectionName, 'Width'         , Width                 );
          H              :=                ReadInteger( aSectionName, 'Height'        , Height                );
          WinState       := TWindowState ( ReadInteger( aSectionName, 'WindowState'   , Integer( WindowState)));
          ZoomFactor     :=                ReadFloat  ( aSectionName, 'ZoomFactor'    , ZoomFactor            );
          BottomHeight   :=                ReadInteger( aSectionName, 'BottomHeight'  , BottomHeight          );
          DetailsWidth   :=                ReadInteger( aSectionName, 'DetailsWidth'  , DetailsWidth          );
          AutomationSize :=                ReadInteger( aSectionName, 'AutomationSize', AutomationSize        );
          UseSplines     :=                ReadBool   ( aSectionName, 'UseSplines'    , UseSplines            );
          LockTime       :=                ReadBool   ( aSectionName, 'LockTime'      , LockTime              );
          LockValue      :=                ReadBool   ( aSectionName, 'LockValue'     , LockValue             );

          if ( L < 0) or ( L > Screen.Width  - 20) then L := 0;
          if ( T < 0) or ( T > Screen.Height - 20) then T := 0;
          if W > Screen.Width                      then W := Screen.Width  - 20;
          if H > Screen.Height                     then H := Screen.Height - 20;

          SetBounds( L, T, W, H);

          if WinState = wsMaximized
          then WindowState := wsMaximized;

          for i := Low( FColWidths) to High( FColWidths)
          do FColWidths[ i] := ReadInteger( aSectionName, Format( 'ColWidth%d', [ i], AppLocale), ListView.Columns[ i].Width)
        end;
      end;
    end;


    procedure   TFormAutomation.FillStatusFeedback( const aLabel: TLabel);
    begin
      if  Assigned( aLabel    )
      and Assigned( FTimeIndex)
      then begin
        if Playing
        then begin
          if Paused
          then begin
            aLabel.Caption    := Format( 'Play ... [%d / %d]'      , [ PlayPointer + 1, FTimeIndex.Count], AppLocale);
            aLabel.Font.Color := clGreen;
          end
          else begin
            aLabel.Caption    := Format( 'Play[%d / %d]'           , [ PlayPointer + 1, FTimeIndex.Count], AppLocale);
            aLabel.Font.Color := clLime;
          end;
        end
        else if Recording
        then begin
          if Paused
          then begin
            aLabel.Caption    := Format( 'Record ... [%d / %d]'    , [ FillCount, Size], AppLocale);
            aLabel.Font.Color := clMaroon;
          end
          else begin
            aLabel.Caption    := Format( 'Record [%d / %d]'        , [ FillCount, Size], AppLocale);
            aLabel.Font.Color := clRed;
          end;
        end
        else begin
          aLabel.Caption    := Format( 'Stopped [%d / %d]'         , [ FillCount, Size], AppLocale);
          aLabel.Font.Color := clBlack;
        end;
      end;
    end;


    procedure   TFormAutomation.FillLedFeedback( const aLed: TKnobsLed);
    begin
      if  Assigned( aLed      )
      and Assigned( FTimeIndex)
      then begin
        if Playing
        then begin
          if Paused
          then begin
            aLed  .Active  := True;
            aLed  .ColorOn := clGreen;
            aLed  .Hint    := 'Play pause';
          end
          else begin
            aLed  .Active  := True;
            aLed  .ColorOn := clLime;
            aLed  .Hint    := 'Playing';
          end;
        end
        else if Recording
        then begin
          if Paused
          then begin
            aLed  .Active  := True;
            aLed  .ColorOn := clMaroon;
            aLed  .Hint    := 'Rec. pause';
          end
          else begin
            aLed  .Active  := True;
            aLed  .ColorOn := clRed;
            aLed  .Hint    := 'Recording';
          end;
        end
        else begin
          aLed  .Active := False;
          aLed  .Hint   := 'Automation off';
        end;
      end;
    end;


    procedure   TFormAutomation.FlagAutomationChanged;
    begin
      FAutomationChanged := True;
    end;


    procedure   TFormAutomation.FixTransportButtons( aRecordButton, aPlayButton, aPauseButton, aLoopButton: TSpeedButton);
    begin
      if  Assigned( aRecordButton)
      and Assigned( aPlayButton  )
      and Assigned( aPauseButton )
      and Assigned( aLoopButton  )
      then begin
        aLoopButton  .Enabled := True;
        aPauseButton .Enabled := True;
        aPlayButton  .Enabled := True;
        aRecordButton.Enabled := True;
      end;
    end;


    procedure   TFormAutomation.SaveAutomation;
    var
      aStringList : TStringList;
    begin
      if  Assigned( FEditor)
      and Assigned( FAutomationData)
      then begin
        FixTimes;
        aStringList := TStringList.Create;

        try
          FAutomationData.SaveToStringList( aStringList, FEditor.Guid);
          aStringList.SaveToFile( ChangeFileExt( FEditor.Filename, EXT_AUTO));
        finally
          aStringList.DisposeOf;
        end;
      end;
    end;


    procedure   TFormAutomation.LoadAutomation;
    var
      aStringList : TStringList;
    begin
      if  Assigned( FEditor)
      and Assigned( FAutomationData)
      then begin
        aStringList := TStringList.Create;

        try
          aStringList.LoadFromFile( ChangeFileExt( FEditor.Filename, EXT_AUTO));
          FAutomationData.LoadFromStringList( FEditor, aStringList, FEditor.Guid);
          FixAutomation;
          FixGridFromData;
        finally
          aStringList.DisposeOf;
        end;
      end;
    end;


    procedure   TFormAutomation.FixAutomation;
    begin
      // todo : there may be controls in the automation system that have no value set for the first and last time
      //        this makes KnobsDataMaker unhappy - look into this ...
    end;


    procedure   TFormAutomation.Tick( CurrentTime: TKnobsTimeStamp);
    begin
      // todo: Play timed, should be called once every ms and then call Clock() on time

      if  Assigned( FAutomationData)
      and Assigned( FTimeIndex)
      and Playing
      and not Paused
      then begin
        if CurrentTime < FLastTickTime
        then begin
          if CurrentTime = 0
          then FLastTickTime := CurrentTime
          else FLastTickTime := CurrentTime - 1;
        end;

        FTickTimeDelta := CurrentTime - FLastTickTime;
        FLastTickTime  := CurrentTime;

        if FTickTimeDelta > FlargestTickTimeDelta
        then FlargestTickTimeDelta := FTickTimeDelta;
      end;
    end;


    procedure   TFormAutomation.Clock;
    var
      i : Integer;
    // Play all items having the same timestamp as the 'current one'
    begin
      if  Assigned( FAutomationData)
      and Assigned( FTimeIndex)
      and Playing
      and not Paused
      then begin
        if  ( PlayPointer >= -1        )
        and ( PlayPointer <= FUsedWidth)
        then begin
          if Reversed
          then begin
            if  Looping
            and ( PlayPointer <= 0)
            then PlayPointer := FUsedWidth  - 1
            else PlayPointer := PlayPointer - 1
          end
          else begin
            if  Looping
            and ( PlayPointer >= FUsedWidth - 1)
            then PlayPointer := 0
            else PlayPointer := PlayPointer + 1;
          end;

          for i := 0 to FTimeIndex.ItemCount[ PlayPointer] - 1
          do FAutomationData.PlayItem( FTimeIndex.Item[ PlayPointer, i]);
        end
        else begin
          if Looping
          then begin
            if Reversed
            then PlayPointer := FUsedWidth
            else PlayPointer := 0;
          end;
        end
      end;
    end;


// Delphi area

{$R *.dfm}

procedure TFormAutomation.CheckBoxLockTimeClick(Sender: TObject);
begin
  LockTime := CheckBoxLockTime.Checked;
end;

procedure TFormAutomation.CheckBoxLockValueClick(Sender: TObject);
begin
  LockValue := CheckBoxLockValue.Checked;
end;

procedure TFormAutomation.ListViewSelectItem(Sender: TObject; Item: TListItem; Selected: Boolean);
begin
  if Selected
  then ListviewClicked( Item);
end;

procedure TFormAutomation.EditAutomationSizeChange(Sender: TObject);
var
  N : Integer;
begin
  N := StrToIntDef( EditAutomationSize.Text, -1);

  if N >= 0
  then AutomationSize := N;
end;

procedure TFormAutomation.EditDummyChange(Sender: TObject);
begin
  if EditDummy.Text <> ''
  then EditDummy.Text := '';
end;

procedure TFormAutomation.ControlUnfocus(const aSender: TObject);
begin
  EditDummy.SetFocus;
end;

procedure TFormAutomation.datagraph_splinesValueChanged(const aSender: TObject; const aPath, aControlType: string;
  aValue: Double; IsFinal, IsAutomation: Boolean);
begin
  UseSplines := Round( aValue) = 1;
end;

procedure TFormAutomation.BitBtnLoadRecordingClick(Sender: TObject);
begin
  LoadAutomation;
end;

procedure TFormAutomation.BitBtnSaveRecordingClick(Sender: TObject);
begin
  SaveAutomation;
end;

procedure TFormAutomation.BitBtnToggleNamesClick(Sender: TObject);
begin
  UseAlternateTitles := not UseAlternateTitles;
end;

procedure TFormAutomation.KnobsAutomationGridClick(const aSender: TObject; aButton: TMouseButton; aShift: TShiftState;
  anX, anY: Integer);
begin
  CursorPosition := Point( anX, anY);
  FixListView;
end;

procedure TFormAutomation.KnobsDataMakerPointChanged(const aSender: TObject; const aPath: string; anIndex: Integer; anX,
  anY: Double);
begin
  if Assigned( FEditDataTimeIndex)
  then HandleGridPointChanged( aSender, aPath, anIndex, anX, anY);
end;

procedure TFormAutomation.KnobsKnobKnobPositionChanged(const aSender: TKnobsValuedControl; const aPath,
  aControlType: string; aPosition: Integer);
begin
  if Assigned( Associate)
  then Associate.KnobPosition := aPosition;
end;

procedure TFormAutomation.KnobsLedAutomationActiveClick(Sender: TObject);
begin
  Stop;
end;

procedure TFormAutomation.KnobsSelectorSendValueChanged(const aSender: TObject; const aPath, aControlType: string;
  aValue: Double; IsFinal, IsAutomation: Boolean);
begin
  SendValues;
end;

procedure TFormAutomation.KnobsSelectorClearValueChanged(const aSender: TObject; const aPath, aControlType: string;
  aValue: Double; IsFinal, IsAutomation: Boolean);
begin
  ClearAutomation;
end;

procedure TFormAutomation.KnobsSelectorSnapValueChanged(const aSender: TObject; const aPath,
  aControlType: string; aValue: Double; IsFinal, IsAutomation: Boolean);
begin
  TakeSnapShot;
end;

procedure TFormAutomation.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if Assigned( FEditor)
  then FEditor.SetActiveInDesigner( '', True);
end;

procedure TFormAutomation.FormCreate(Sender: TObject);
begin
  FAutomationData                := TKnobsEditHistory.Create( nil, 1, HISTORY_STR_LEN, True );
  FTimeIndex                     := TKnobsTimeIndex.Create;
  FEditDataTimeIndex             := TKnobsEditDataTimeIndex.Create;
  FZoomFactor                    := Sqrt( 2); ZoomFactor := 1.0;
  FCursorPosition                := Point(  0,  0);
  CursorPosition                 := Point( -1, -1);
  KnobsAutomationGrid.Left       := 0;
  KnobsAutomationGrid.Top        := 0;
  FAutomationData.AllowNonAuto   := False; // do not allow non automatable controls to end up in the data
  FAutomationData.AllowUnchanged := True;  // do     allow unchanged       controls to end up in the data
  AutomationSize                 := DEFAULT_AUTOMATION_SIZE;
  FAutomationChanged             := True;
  FLockTime                      := False;
  LockTime                       := True;
  FLockValue                     := True;
  LockValue                      := False;
end;

procedure TFormAutomation.FormDestroy(Sender: TObject);
begin
  UnregisterCallbacks;
  FreeAndNil( FEditDataTimeIndex);
  FreeAndNil( FTimeIndex        );
  FreeAndNil( FAutomationData   );
end;

procedure TFormAutomation.FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if WindowFromPoint( Mouse.CursorPos) = ScrollBoxAutomationGrid.Handle
  then begin
    if Key = VK_UP
    then begin
      if ssShift in Shift
      then GoToUpStop
      else GoUp;

      Key := 0;
    end
    else if Key = VK_RIGHT
    then begin
      if ssShift in Shift
      then GoToRightStop
      else GoRight;

      Key := 0;
    end
    else if Key = VK_DOWN
    then begin
      if ssShift in Shift
      then GoToDownStop
      else GoDown;

      Key := 0;
    end
    else if Key = VK_LEFT
    then begin
      if ssShift in Shift
      then GoToLeftStop
      else GoLeft;

      Key := 0;
    end
  end;
end;

procedure TFormAutomation.FormMouseWheel(Sender: TObject; Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint;
  var Handled: Boolean);
var
  aMsg  : Cardinal;
  aCode : Cardinal;
  i     : Integer;
  n     : Integer;
begin
  if WindowFromPoint( Mouse.CursorPos) = ScrollBoxAutomationGrid.Handle
  then begin
    Handled := True;

    if ssShift in Shift
    then begin
      aMsg := WM_HSCROLL;

      if WheelDelta < 0
      then aCode := SB_LINEUP
      else aCode := SB_LINEDOWN;
    end
    else begin
      aMsg := WM_VSCROLL;

      if WheelDelta > 0
      then aCode := SB_LINEUP
      else aCode := SB_LINEDOWN;
    end;

    n := Mouse.WheelScrollLines;

    for i := 0 to n
    do ScrollBoxAutomationGrid.Perform( aMsg, aCode, 0);

    ScrollBoxAutomationGrid.Perform( aMsg, SB_ENDSCROLL, 0);
  end;
end;

procedure TFormAutomation.FormResize(Sender: TObject);
begin
  EditDummy.Left := Width + 10;
end;

procedure TFormAutomation.FormShow(Sender: TObject);
begin
  FAutomationChanged := True;
  FShown             := True;
  EditDummy.SetFocus;
  FixButtonEnables;
  KnobsDataMaker.AllowXMoves := not LockTime;
  KnobsDataMaker.AllowYMoves := not LockValue;

  if Reversed
  then PlayPointer := FUsedWidth
  else PlayPointer := -1;
end;

procedure TFormAutomation.TimerBlinkTimer(Sender: TObject);
begin
  BlinkUpdate;
end;

procedure TFormAutomation.TimerStatusUpdateTimer(Sender: TObject);
begin
  StatusUpdate;
end;

procedure TFormAutomation.SpeedButtonLeftFastClick(Sender: TObject);
begin
  GoLeft;
end;

procedure TFormAutomation.SpeedButtonLeftStopClick(Sender: TObject);
begin
  GoToLeftStop;
end;

procedure TFormAutomation.SpeedButtonLoopClick(Sender: TObject);
begin
  Looping := SpeedButtonLoop.Down;
end;

procedure TFormAutomation.SpeedButtonPauseClick(Sender: TObject);
begin
  Paused := SpeedButtonPause.Down;
end;

procedure TFormAutomation.SpeedButtonPlayClick(Sender: TObject);
begin
  Playing := SpeedButtonPlay.Down;
end;

procedure TFormAutomation.SpeedButtonRecordClick(Sender: TObject);
begin
  Recording := SpeedButtonRecord.Down;
end;

procedure TFormAutomation.SpeedButtonRightFastClick(Sender: TObject);
begin
  GoRight;
end;

procedure TFormAutomation.SpeedButtonRightStopClick(Sender: TObject);
begin
  GoToRightStop;
end;

procedure TFormAutomation.SpeedButtonStopClick(Sender: TObject);
begin
  Stop;
end;

procedure TFormAutomation.SpinButtonZoomDownClick(Sender: TObject);
begin
  LessZoom;
end;

procedure TFormAutomation.SpinButtonZoomUpClick(Sender: TObject);
begin
  MoreZoom;
end;

end.

